import re
import os

def preprocessing(text: str, offset: [tuple, list]):
    
    assert len(offset) == 2
    off1, off2 = [int(el) for el in offset]

    text = text[:off1] + "< < " + re.sub("\s+", " > > < < ", text[off1:off2].strip()) + " > > " + text[off2:]
    text = re.sub("\s+", " ", text.strip()).lower()
    
    return text

current_dir = os.path.dirname(os.path.realpath(__file__))


sorted_content_termins_twoplus = []
with open(os.path.join(current_dir, 'data', 'complex_content_termins.txt')) as f:
    for l in f.readlines():
        sorted_content_termins_twoplus.append(l.strip())
        
sorted_grammar_termins_twoplus = []
with open( os.path.join(current_dir, 'data', 'complex_grammar_termins.txt')) as f:
    for l in f.readlines():
        sorted_grammar_termins_twoplus.append(l.strip())

def adjust_grammar_terms(string):
    
    string = re.sub('(?<!\b)[ ](?=>)','',string)
    
    string = re.sub(' - ','-',string)
    string = re.sub(' \+ ','+',string)
    
    used_terms = []
    for gt in sorted_grammar_termins_twoplus:
        if len(re.findall(f"[ ](?={gt}>)",string))>0:
            used_terms.append(gt)
            string = re.sub(f"[ ](?={gt}>(?!>))"," <",string)

    open_brack = False
    new_list = []
    for t in string.split():
        if ">" in t and '>>' not in t:
            if open_brack == False:
                t = '<' + t
                new_list.append(t)
                open_brack = False
                continue
            open_brack = False
        elif '<' in t and '<<' not in t:
            open_brack = True
        new_list.append(t)
    
    return ' '.join(new_list)


def adjust_content_terms(string, verbose = False):
#     used_terms = []
    for ct in sorted_content_termins_twoplus:  

        string, subcount = re.subn(f"[ ](?={ct}>>)"," <<",string)
        if subcount > 0:
            if verbose == True: print(ct)
#             used_terms.append(ct)
            
    open_brack = False
    new_list = []
    for t in string.split():
        if '>>' in t:
            if open_brack == False:
                t = '<<' + t
                new_list.append(t)
                open_brack = False
                continue
            open_brack = False
        elif  '<<' in t:
            open_brack = True
        new_list.append(t)

    return ' '.join(new_list)


def upper_repl(match):
     return match.group(0).upper() 
    
    
def post_process_string(model_gener_string):
    adj_string = adjust_grammar_terms(model_gener_string.strip())
    adj_string = adjust_content_terms(adj_string)
    adj_string = adj_string[0].upper() + adj_string[1:]
    adj_string = re.sub(r"(?<=\. )\w", upper_repl, adj_string)
    return adj_string


from transformers import T5ForConditionalGeneration, AutoTokenizer
import torch

from .processing import preprocessing, post_process_string

class FeedbackGenerator():
    
    def __init__(self, cuda_index: [bool, int, str], model_name = "SkolkovoInstitute/GenChal_2022_nigula"):
        
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = T5ForConditionalGeneration.from_pretrained(model_name)
        self.model.eval();
        
        if cuda_index != False:
            device = torch.device(f'cuda:{cuda_index}')
            self.model.to(device);
        
        
    def get_feedback(self, text, error_offsets:list, temperature=1.0, beams=3, 
                   num_return_sequences = 1, do_sample = False,
                  repetition_penalty = 1.1, max_length_multiplier = 3):
        texts = [text] if isinstance(text, str) else text
        assert len(error_offsets) == len(texts)
        
        texts_processed = []
        for txt, off in zip(texts,error_offsets):
            txt_processed = preprocessing(txt, off)
            texts_processed.append(txt_processed)        
                    
        inputs = self.tokenizer(texts_processed, return_tensors='pt', padding=True)['input_ids'].to(self.model.device)
        result = self.model.generate(
            inputs, 
            num_return_sequences=num_return_sequences, 
            do_sample=do_sample, 
            temperature=temperature, 
            repetition_penalty=repetition_penalty, 
            max_length=int(inputs.shape[1] * max_length_multiplier) ,
            num_beams=beams,
        )
        texts = [self.tokenizer.decode(r, skip_special_tokens=True) for r in result]
                
        if isinstance(text, str):
            return post_process_string(texts[0])
        
        texts = [post_process_string(t) for t in texts]
        
        return texts


import spacy
try:
    nlp=spacy.load('en_core_web_md')
except Exception as e:
    import os
    print(os.popen('python3 -m spacy download en_core_web_md').read())
    nlp=spacy.load('en_core_web_md')
    
    # raise Exception(f"Failed with {e}, try running the following command - < python3 -m spacy download en_core_web_md > and reload your notebook")
    
from transformers import pipeline

def isbetween(index, offset_list):
    return offset_list[0] <= int(index) <= offset_list[1]

def get_smart_truncated_string(text, err_word_offset, debug = False):
    
    spacy_parsing = nlp(text)
    err_word = text[err_word_offset[0]:err_word_offset[1]]
    
    text = text.lower()
    text_keep_initial = text[:err_word_offset[1]]
    err_word = err_word.lower()
    
    if debug == True:
        print(text)
        print(err_word_offset)
        print(f"error word <{err_word}>")
    
    words_to_keep = err_word.split()
    last_index_to_keep = err_word_offset[1]

    error_syntax_pairs = []
 
    for token in spacy_parsing:
        # print("{:<15} {:^10} {:>15} {:>15}".format(str(token.head.text), str(token.dep_), str(token.text), str(token.head.idx)))
        if token.dep_ != 'punct':
            if isbetween(token.head.idx, err_word_offset) or isbetween(token.idx, err_word_offset):
                if debug == True:print("{:<15} {:^10} {:>15} {:>15}  {:>15}".format(str(token.head.text), str(token.dep_), str(token.text), str(token.idx), str(token.head.idx)))

                words_to_keep.append(token.head.text)
                words_to_keep.append(token.text)
                
                end_of_token = token.idx + len(token.text)
                end_of_token_head = token.head.idx + len(token.head.text)
                maximal_end = max(end_of_token, end_of_token_head)
                
                pair = token.head.text + '_' + token.dep_ + '_' + token.text
                error_syntax_pairs.append(pair)

                if maximal_end > last_index_to_keep:
                    last_index_to_keep = maximal_end
                    if debug == True:print("Enlarge index", maximal_end)

            else:
                # print(f"{token.head.idx} and {token.idx} not between {err_word_offset}")
                pass
        
        
    words_to_keep = list(set(words_to_keep))
    
    
    THRESHOLD = 3
    #if there are not enough connections engage more words
    if len(words_to_keep) - len(err_word.split()) <=THRESHOLD:
        for token in spacy_parsing:
            if token.dep_ != 'punct':
                if token.head.text in words_to_keep or token.head.text in words_to_keep:

                    if any([t not in words_to_keep for t in [token.head.text, token.text]]):
                        
                        if debug == True:print("{:<15} {:^10} {:>15} {:>15}  {:>15}".format(str(token.head.text), str(token.dep_), str(token.text), str(token.idx), str(token.head.idx)))
                        
                        token_to_process = token.head if token.head.text not in words_to_keep  else  token

                        end_of_token = token_to_process.idx + len(token_to_process.text)
                        
                        words_to_keep.append(token_to_process.text)
                        
                        pair = token.head.text + '_' + token.dep_ + '_' + token.text
                        error_syntax_pairs.append(pair)

                        if end_of_token > last_index_to_keep:
                            
                            last_index_to_keep = end_of_token
                            
                            if debug == True:print("Enlarge index by force", end_of_token)
                        else:
                            if debug == True:print(f"Force adding of word <{token_to_process.text}>")
 
            if len(words_to_keep)> THRESHOLD:# if the dependent word are already available in the back then it is ok and we stop
                break  
        
    text_to_keep = text[:last_index_to_keep]
    words_to_keep = '_'.join(sorted(set(words_to_keep))).lower()
    
    error_syntax_pairs = sorted([el.lower() for el in error_syntax_pairs if el.split('_')[0] != el.split('_')[1]])
    error_syntax_pairs = '&'.join(error_syntax_pairs)
    
    if debug == True:
        print(text_keep_initial)
        print(">>>")
        print(text_to_keep)
        
        print(f"words_to_keep {words_to_keep}")
        
        print(f"error_syntax_pairs {error_syntax_pairs}")
        
    return text_to_keep


class Augmenter():
    
    def __init__(self, cuda_index: [int, str], model_name = "EleutherAI/gpt-neo-1.3B"):
        
        self.generator = pipeline('text-generation', model=model_name, device = cuda_index)
        
    def augment(self, text: str, is_text_truncated: [list, tuple], err_word_offset = None,
               min_length = 20, max_length = 100, do_sample = True):
        
        if is_text_truncated == False:
            assert err_word_offset is not None
            text = get_smart_truncated_string(text,err_word_offset)
            
        new_texts = self.generator(text, do_sample=do_sample, min_length=min_length, max_length = max_length)
        
        return new_texts